/* -----------------------------------------------------------------------------

  1996 Dietmar Eilert

  This example has been written to demonstrate how a programmer could add
  GoldED support to a source level debugger. Dice:

  dcc main.c -// -proto -mRR -mi -pr -2.0 -o ram:debugger

  1. Simulate compiler pass: create random breakpoints for testfile "golded:developer/examples/api/mirror/funcs.c"
  2. Launch editor
  3. Open debugger window
  4. Advice GoldED to start debugger session
  5. Show PC (current line of execution)
  6. Allow user to activate/deactivate breakpoints
  7. Run until the debugger window is closed
  8. Advice GoldED to stop debugger session

  ------------------------------------------------------------------------------
*/

/// "includes"

#include <exec/exec.h>
#include <string.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <intuition/intuition.h>
#include <dos/dos.h>
#include <dos/dosextens.h>
#include <dos/rdargs.h>
#include <dos/dostags.h>
#include <workbench/startup.h>
#include <workbench/workbench.h>
#include <utility/tagitem.h>
#include <rexx/errors.h>
#include <rexx/rxslib.h>

#include <clib/alib_protos.h>
#include <clib/dos_protos.h>
#include <clib/exec_protos.h>
#include <clib/icon_protos.h>
#include <clib/intuition_protos.h>
#include <clib/utility_protos.h>
#include <clib/rexxsyslib_protos.h>
#include <clib/wb_protos.h>

#define Prototype extern

///
/// "globals"

ULONG RexxOutstanding;                               // counter of sent messages
UWORD Breakpoints[20];                               // random array of breakpoints

///
/// "prototypes"

Prototype int    wbmain(struct WBStartup *);
Prototype void   main(ULONG, UBYTE **);
Prototype void   Message(UBYTE *);
Prototype void   Demo(struct MsgPort *);
Prototype void   FreeRexxMessage  (struct RexxMsg *);
Prototype void   HandleGoldEDQuery(struct RexxMsg *, struct MsgPort *);
Prototype void   BreakpointsDemo(struct MsgPort *, UBYTE *, struct Window *);
Prototype void   StopSession    (struct MsgPort *, UBYTE *, struct Window *);
Prototype ULONG  ShowPC         (struct MsgPort *, UBYTE *, struct Window *);
Prototype ULONG  QueryBreakpointsHandler(ULONG *, struct MsgPort *);
Prototype ULONG  SetBreakpointHandler   (ULONG *, struct MsgPort *);
Prototype ULONG  SendRexxCommand(UBYTE *, UBYTE *, struct MsgPort *, BOOL);
Prototype UBYTE *xsprintf(UBYTE *, APTR);
Prototype BOOL   VerifyTestfile(void);
Prototype void   InitBreakpointArray(UWORD *);
Prototype UBYTE *StartGoldED(BOOL);

Prototype struct Window *OpenDebuggerWindow(void);
Prototype struct Window *StartSession(struct MsgPort *, UBYTE *);

///
/// "entry points"

/* --------------------------------------- main --------------------------------

 CLI entry point: We open our port DEBUGGER (used to send/recieve messages) and
 start the demo.

*/

void
main(argc, argv)

ULONG argc;
UBYTE *argv[];
{
    struct MsgPort *port;

    if (VerifyTestfile()) {

        // create our message port (used to recieve and send messages)

        if (port = CreateMsgPort()) {

            port->mp_Node.ln_Name = "DEBUGGER";
            port->mp_Node.ln_Pri  =  0;

            AddPort(port);

            Demo(port);

            RemPort(port);

            DeleteMsgPort(port);
        }
    }
}

/* ------------------------------------ wbmain ---------------------------------

 Workbench entry point.

*/

int
wbmain(struct WBStartup *wbs)
{
    return(main(0, NULL));
}

///
/// "demo"

/* ------------------------------------- Demo ----------------------------------

 Run demo

*/

void
Demo(ourPort)

struct MsgPort *ourPort;
{
    UBYTE *golded;

    // simulate compiler pass (create random breakpoints array)

    InitBreakpointArray(Breakpoints);

    // start editor

    if (golded = StartGoldED(TRUE)) {

        struct Window *win;

        // start debugger session

        if (win = StartSession(ourPort, golded)) {

            // demo: how to show PC (current line of execution)

            if (ShowPC(ourPort, golded, win) == RC_OK) {

                // demo: how to handle breakpoint (de)activation

                BreakpointsDemo(ourPort, golded, win);
            }

            // terminate debugger session

            StopSession(ourPort, golded, win);
        }
    }
}


/* ------------------------------ VerifyTestfile -------------------------------

 Check if example source file exists.

*/

BOOL
VerifyTestfile()
{
    BPTR lock;

    if (lock = Lock("golded:developer/examples/api/mirror/funcs.c", ACCESS_READ))
        UnLock(lock);
    else
        Message("Please create example source code golded:developer/examples/api/mirror/funcs.c");

    return(lock != NULL);
}


/* ----------------------------------- StartGoldED -----------------------------

 Run GoldED (check for running GoldED tasks first). Return name of GoldED's
 ARexx port.

*/

UBYTE *
StartGoldED(hide)

BOOL hide;
{
    static UBYTE host   [255];
    static UBYTE command[255];

    strcpy(host, "GOLDED.1");

    // look for running GoldED

    Forbid();

    while (host[7] <= '9') {

        if (FindPort(host)) {

            Permit();

            return(host);
        }
        else
            host[7] += 1;
    }

    Permit();

    // launch GoldED

    strcpy(command, "GOLDED:GOLDED");

    if (hide)
        strcat(command, " HIDE");

    if (SystemTags(command, SYS_Asynch, TRUE, SYS_Input, NULL, SYS_Output, NULL, NP_StackSize, 8192, TAG_DONE) == 0) {

        UWORD try;

        for (try = 50; try; try--, Delay(10))

            if (FindPort("GOLDED.1"))

                return("GOLDED.1");
    }

    return(NULL);
}


/* ----------------------------- InitBreakpointArray ---------------------------

 Create random array of breakpoints (simulate compiler pass)

*/

void
InitBreakpointArray(array)

UWORD *array;
{
    // create random breakpoints (1 = inactive, 2 = active)

    array[4 ] = 1;                                   // line 5
    array[14] = 2;                                   // line 15
    array[17] = 1;                                   // line 18
}


/* ------------------------------- StartSession --------------------------------

 Start a debugger session. GoldED will start asking us for breakpoint lists soon
 at our port for every file loaded into the editor until the session's end.

*/

struct Window *win;
StartSession(ourPort, golded)

struct MsgPort *ourPort;
UBYTE          *golded;
{
    struct Window *win = NULL;

    // start session

    if (SendRexxCommand(golded, "DEBUG START PORT=DEBUGGER", ourPort, FALSE) == RC_OK)

        win = OpenDebuggerWindow();

    return(win);
}


/* ---------------------------------- ShowPC -----------------------------------

 Show line of execution. Randomly choosen by this demo to be line 4 (first line
 of text = line 0).

*/

ULONG
ShowPC(ourPort, golded, win)

struct MsgPort *ourPort;
struct Window  *win;
UBYTE          *golded;
{
    return(SendRexxCommand(golded, "PC FILE=\"golded:developer/examples/api/mirror/funcs.c\" LOCKED LINE=5 COLUMN=1 HIGHLIGHT", ourPort, FALSE));
}


/* ---------------------------- OpenDebuggerWindow -----------------------------

 Open debugger window

*/

struct Window *
OpenDebuggerWindow()
{
    struct Window *win;

    win = OpenWindowTags(NULL,

        WA_Width,       320,
        WA_Height,      100,
        WA_Title,       "Debugger demo - close window to stop program",
        WA_ScreenTitle, "Debugger demo - close window to stop program",
        WA_Flags,       WFLG_DRAGBAR | WFLG_DEPTHGADGET | WFLG_CLOSEGADGET | WFLG_SMART_REFRESH | WFLG_ACTIVATE | WFLG_RMBTRAP,
        WA_IDCMP,       IDCMP_CLOSEWINDOW,
        TAG_DONE
    );

    return(win);
}


/* ------------------------------ BreakpointsDemo ------------------------------

 Wait for breakpoint activation/deactivation (notififactions sent by GoldED)
 until the debugger window is closed.

*/

void
BreakpointsDemo(ourPort, golded, win)

struct MsgPort *ourPort;
struct Window  *win;
UBYTE          *golded;
{
    ULONG signalMask = 1L<<(ourPort->mp_SigBit) | 1L<<(win->UserPort->mp_SigBit);

    BOOL terminated = FALSE;

    do {

        struct Message *msg;

        Wait(signalMask);

        while (msg = GetMsg(ourPort)) {

            if (((struct RexxMsg *)msg)->rm_Node.mn_Node.ln_Type == NT_REPLYMSG) {

                // one of our async messages has been replied

                --RexxOutstanding;

                FreeRexxMessage((struct RexxMsg *)msg);
            }
            else {

                // somebody (GoldED ?) sends us a message

                HandleGoldEDQuery((struct Rexxmsg *)msg, ourPort);

                ReplyMsg(msg);
            }
        }

        while (msg = GetMsg(win->UserPort)) {

            if (((struct IntuiMessage *)msg)->Class == IDCMP_CLOSEWINDOW)
                terminated = TRUE;

            ReplyMsg(msg);
        }

    } while (terminated == FALSE);
}


/* --------------------------------- StopSession -------------------------------

 Stop debugger session, close debugger window.

*/

void
StopSession(ourPort, golded, win)

struct MsgPort *ourPort;
struct Window  *win;
UBYTE          *golded;
{
    SendRexxCommand(golded, "DEBUG STOP PORT=DEBUGGER", ourPort, TRUE);

    // free breakpoints

    SendRexxCommand(golded, "BREAKPT CLEANUP", ourPort, TRUE);

    // close debugger window

    CloseWindow(win);

    // have all of our messages sent to GoldED been replied ?

    while (RexxOutstanding) {

        struct RexxMsg *msg;

        while (msg = (struct RexxMsg *)GetMsg(ourPort)) {

            if (msg->rm_Node.mn_Node.ln_Type == NT_REPLYMSG) {

                // one of our messages has been replied

                --RexxOutstanding;

                FreeRexxMessage(msg);
            }
            else
                ReplyMsg((struct Message *)msg);
        }

        if (RexxOutstanding)

            WaitPort(ourPort);
    }   
}

///
/// "query handling"


/* ----------------------------- HandleGoldEDQuery -----------------------------

 Somebody (GoldED ?) sends us a message. Parse command, run command handler.

*/

void
HandleGoldEDQuery(rexxMsg, ourPort)

struct RexxMsg *rexxMsg;
struct MsgPort *ourPort;
{
    struct RDArgs *rdArgs, *args;

    ULONG result = RC_OK;

    if (rdArgs = AllocDosObject(DOS_RDARGS, NULL)) {

        // table of supported commands, handlers & template strings

        static struct parser { UBYTE *command;  ULONG (*handler)(ULONG *, struct MsgPort *); UBYTE *template; } parser[] = {

            "QUERY",   (APTR)QueryBreakpointsHandler,  "FILE/K/A,PORT/K/A,DATA/N,ELEMENTS/N",
            "BREAKPT", (APTR)SetBreakpointHandler,     "FILE/K,LINE/N,STATE/N",
             NULL
        };

        ULONG argArray[] = { 0, 0, 0, 0, 0, 0, 0, 0, 0 };

        UBYTE buffer[255];
        UWORD n;

        strncpy(buffer, ARG0(rexxMsg), sizeof(buffer) - 1);

        // make LF-terminated copy of command string (required by dos/readArgs):

        strcat(buffer, "\12");

        for (n = 0; parser[n].command; ++n) {

            if (memcmp(buffer, parser[n].command, strlen(parser[n].command)) == 0) {

                UBYTE *arguments = buffer + strlen(parser[n].command);

                rdArgs->RDA_Source.CS_Buffer = arguments;
                rdArgs->RDA_Source.CS_Length = strlen(arguments);
                rdArgs->RDA_Source.CS_CurChr = 0;
                rdArgs->RDA_DAList           = NULL;
                rdArgs->RDA_Buffer           = NULL;

                if (parser[n].template) {

                    if (args = ReadArgs(parser[n].template, argArray, rdArgs)) {

                        result = (*parser[n].handler)(argArray, ourPort);

                        FreeArgs(args);
                    }
                    else {

                        static UBYTE errorText[80 + 1];

                        Fault(IoErr(), "IoErr()", errorText, 80);

                        Message(errorText);

                        result = RC_ERROR;
                    }
                }
                else
                    result = (*parser[n].handler)(argArray, ourPort);
            }
        }

        FreeDosObject(DOS_RDARGS, rdArgs);
    }

    if (rexxMsg->rm_Action & RXFF_RESULT) {

        rexxMsg->rm_Result1 = result;
        rexxMsg->rm_Result2 = NULL;
    }
}


/* --------------------------- SetBreakpointHandler ----------------------------

 Handle notification sent by GoldED: a breakpoint has been set/cleared/toggled.
 This demo simply flashes the screen. A real debugger would have to do its own
 magic (e.g. remember the breakpoit state).

 argArray[] template: FILE/K,LINE/N,STATE/N

*/

ULONG
SetBreakpointHandler(argArray, ourPort)

ULONG          *argArray;
struct MsgPort *ourPort;
{
    UBYTE *file  =  (UBYTE *)argArray[0];
    ULONG  line  = *(ULONG *)argArray[1];
    ULONG  state = *(ULONG *)argArray[2];

    DisplayBeep(0);

    return(RC_OK);
}


/* -------------------------- QueryBreakpointsHandler --------------------------

 Handle QUERY command sent by GoldED;  Return a random list of breakpoints if
 GoldED asks for the breakpoints of our testfile.

 argArray[] template: FILE/K/A,PORT/K/A,DATA/N,ELEMENTS/N

*/

ULONG
QueryBreakpointsHandler(argArray, ourPort)

ULONG          *argArray;
struct MsgPort *ourPort;
{
    BPTR lockA, lockB;

    lockA = Lock((UBYTE *)argArray[0], ACCESS_READ);

    lockB = Lock("golded:developer/examples/api/mirror/funcs.c", ACCESS_READ);

    // print list of breakpoints set by user before debug start (ignored)

    if (argArray[2] && argArray[3]) {

        ULONG  elements, data, line;
        UWORD *array;

        elements = *(ULONG *)argArray[3];
        data     = *(ULONG *)argArray[2];

        array = (UWORD *)data;

        for (line = 1; elements--; ++line) {

            switch (*array++) {

                case 1:

                    printf("breakpoint in line %d (unused)\n", line);
                    break;

                case 2:

                    printf("breakpoint in line %d (set)\n", line);
                    break;
            }
        }
    }

    // breakpoints of testfile wanted ?

    if (lockA && lockB && (SameLock(lockA, lockB) == LOCK_SAME)) {

        static UBYTE command[255];

        sprintf(command, "BREAKPT FILE=\"%s\" FORMAT=STORMC DATA=%ld ELEMENTS=%ld", (UBYTE *)argArray[0], Breakpoints, sizeof(Breakpoints) / sizeof(UWORD));

        // send list of breakpoints

        SendRexxCommand((UBYTE *)argArray[1], command, ourPort, TRUE);
    }

    if (lockA)
        UnLock(lockA);

    if (lockB)
        UnLock(lockB);

    return(RC_OK);
}

///
/// "misc"

/* ----------------------------------- Info ------------------------------------

 Display info requester

*/

void
Message(info)

UBYTE *info;
{   
    struct EasyStruct easyStruct = { sizeof(struct EasyStruct), 0, "Info", info, "OK" };

    EasyRequestArgs(NULL, &easyStruct, NULL, NULL);
}


/* --------------------------------- xsprintf ----------------------------------

 sprintf frontend (returns pointer to static buffer)

*/

UBYTE *
xsprintf(template, data)

UBYTE *template;
APTR  data;
{
    static UBYTE buffer[255];

    return(sprintf(buffer, template, data));
}


/* ---------------------------------- SendRexxCommand -------------------------

 <async> = TRUE:

 Send ARexx message, return immediately. Increment global counter of outstanding
 messages (messages that have not yet been replied).

 <async> = NULL:

 Send ARexx message & wait for answer. Return result code. This is a blocking
 function but it can handle incoming commands sent by GoldED while waiting for
 a reply, thus avoiding deadlocks (waiting for GoldED to reply our message
 while GoldED is waiting for us to reply one of its messages).

*/

ULONG
SendRexxCommand(port, cmd, ourPort, async)

struct MsgPort *ourPort;
UBYTE          *cmd, *port;
BOOL            async;
{
    struct MsgPort *rexxport;

    ULONG result = RC_ERROR;

    Forbid();

    if (rexxport = FindPort(port)) {

        struct RexxMsg *rexxMsg, *answer;

        if (rexxMsg = CreateRexxMsg(ourPort, NULL, NULL)) {

            if (rexxMsg->rm_Args[0] = CreateArgstring(cmd, strlen(cmd))) {

                rexxMsg->rm_Action = RXCOMM | RXFF_RESULT;

                PutMsg(rexxport, &rexxMsg->rm_Node);

                Permit();

                if (async) {

                    // async == TRUE: return immediately

                    ++RexxOutstanding;

                    result = RC_OK;
                }
                else {

                    // async == FALSE: wait for reply

                    for (;;) {

                        if (answer = (struct RexxMsg *)GetMsg(ourPort)) {

                            if (answer == rexxMsg) {

                                // done: our message has been replied

                                result = answer->rm_Result1;

                                FreeRexxMessage(answer);

                                break;
                            }
                            else if (answer->rm_Node.mn_Node.ln_Type == NT_REPLYMSG) {

                                // one of our async messages has been replied

                                --RexxOutstanding;

                                FreeRexxMessage(answer);
                            }
                            else {

                                // somebody (GoldED ?) sends us a message

                                HandleGoldEDQuery(answer, ourPort);

                                ReplyMsg((struct Message *)answer);
                            }
                        }

                        WaitPort(ourPort);
                    }
                }
            }
        }
        else
            Permit();
    }
    else
        Permit();

    return(result);
}


/* ------------------------------ FreeRexxMessage ------------------------------

 Free one of our RexxMessages (replied by reciever already)

*/

void
FreeRexxMessage(rexxMsg)

struct RexxMsg *rexxMsg;
{
    DeleteArgstring((UBYTE *)ARG0(rexxMsg));

    DeleteRexxMsg(rexxMsg);
}

///
